import numpy as np
import numba as nb
import networkx as nx
import itertools
from scipy.cluster.vq import vq
from .graphs import ConstructGraph, GetSubGraph
from .core import (
    PartitionData,
    # PartitionData_cp,
    Encode2ElasticMatrix,
    DecodeElasticMatrix,
)


#####################################
### Time supervision utils ###
#####################################


# -----extract oriented branches and associated data
def bf_search(dict_branches, root_node):
    """ breadth-first tree search: returns edges and nodes ordering for dict_branches given root_node"""
    flat_tree = nx.Graph()
    flat_tree.add_nodes_from(
        list(set(itertools.chain.from_iterable(dict_branches.keys())))
    )
    flat_tree.add_edges_from(dict_branches.keys())
    edges = list(nx.bfs_edges(flat_tree, root_node))
    nodes = [root_node] + [v for u, v in edges]
    return edges, nodes


def get_circle_or_curve(Edges, root_node):
    """Given root node, decompose Edges into graph relations: circle or curve version """
    # ----find ordered relations from root node (for the single "branch")
    _dict_branches = {tuple(x): None for x in Edges}
    ordered_edges, ordered_nodes = bf_search(_dict_branches, root_node)
    dict_branches_single_end = {(ordered_nodes[0], ordered_nodes[-1]): ordered_nodes}
    return dict_branches_single_end


def get_tree(Edges, root_node):
    """Given root node, decompose Edges into graph relations:
    - tree parent-children relations
    - branches with (repeating) endpoints,
    - partitioning of nodes into branches ('single-ended')"""
    # ----get branches
    net = ConstructGraph({"Edges": [Edges]})
    branches = GetSubGraph(net, "branches")
    _dict_branches = {
        (b[0], b[-1]): b for i, b in enumerate(branches)
    }  # temporary branch node lists (not in order)

    # ----check validity of the root node
    root_branch = [k for k in _dict_branches.keys() if root_node in k]
    if len(root_branch) > 1:
        raise ValueError(f"Multiple root branches {root_branch}")

    # ----reorder branches
    # find ordered relations between branches
    ordered_edges, ordered_nodes = bf_search(_dict_branches, root_node)

    # ----create ordered dicts
    dict_tree = {}  # branch parent-child relations
    dict_branches = {}  # branch node lists
    dict_branches_single_end = (
        {}
    )  # branch node lists with no overlapping terminal nodes
    # visited_nodes = []
    for i, e in enumerate(ordered_edges):  # for each branch
        # store branch in order (both the key and the list)
        if e not in _dict_branches:
            dict_branches[e] = _dict_branches[e[::-1]][::-1]
        else:
            dict_branches[e] = _dict_branches[e]

        # store children
        dict_tree[e] = [
            _e for _e in ordered_edges[:i] + ordered_edges[i + 1 :] if e[-1] in _e
        ]

        # store single ended branch
        if i == 0:
            dict_branches_single_end[e] = dict_branches[e]  # if n not in visited_nodes]
        else:
            dict_branches_single_end[e] = dict_branches[e][1:]
        # dict_branches_single_end[e] = [n for n in dict_branches[e] if n not in visited_nodes]
        # visited_nodes.extend(dict_branches[e])
    return dict_tree, dict_branches, dict_branches_single_end


def get_shortest_paths(Edges, root_node):
    """ Return paths from root node to branches endpoints """
    # ---make graph
    net = ConstructGraph({"Edges": [Edges]})
    degrees = net.degree()

    # ---check validity of the root node
    if degrees[root_node] > 1:
        raise ValueError("root node has degree > 1")

    # ---get paths
    end_points = [v for v, d in enumerate(degrees) if d == 1 and v != root_node]
    paths = net.get_shortest_paths(root_node, end_points)
    paths = {(p[0], p[-1]): p for p in paths}
    return paths


def partition_data_by_branch(X, SquaredX, NodePositions, branches, TrimmingRadius):
    """Partition data into each branch """
    partition, dists = PartitionData(
        X, NodePositions, 10 ** 8, SquaredX, TrimmingRadius=TrimmingRadius
    )
    branches_dataidx = {k: np.isin(partition[:, 0], b) for k, b in branches.items()}
    return branches_dataidx


def partition_data_by_branch_precomp(partition, branches):
    """Partition data into each branch """
    branches_dataidx = {k: np.isin(partition[:, 0], b) for k, b in branches.items()}
    return branches_dataidx


# ------generate pseudotime centroid branches
def nNodes_pseudotime(bX, bpseudotime, bnNodes):
    """ "
    Argsorts pseudotime and create nNodes bins
    with even number of points  but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    PseudotimeNodePositions = np.zeros((bnNodes, bX.shape[1]))
    MeanPseudotime = np.zeros(bnNodes)
    idx_data = np.linspace(0, len(bX), bnNodes + 1, dtype=int)
    for i in range(bnNodes):
        PseudotimeNodePositions[i] = bX[
            argsort_pseudotime[idx_data[i] : idx_data[i + 1]]
        ].mean(axis=0)
        MeanPseudotime[i] = bpseudotime[
            argsort_pseudotime[idx_data[i] : idx_data[i + 1]]
        ].mean()

    return PseudotimeNodePositions, MeanPseudotime


def bin_pseudotime(bX, bpseudotime, bnNodes):
    """ "
    Argsorts pseudotime and create nNodes bins
    with even amounts of pseudotime but potentially uneven number of points
    """
    # ---create nodes with uniformly spread pseudotime
    count, bins = np.histogram(bpseudotime, bins=bnNodes)
    clusters = np.digitize(bpseudotime, bins[1:], right=True)
    # handle edge case: empty bins
    if 0 in count:
        return nNodes_pseudotime(bX, bpseudotime, bnNodes)

    PseudotimeNodePositions = np.zeros((bnNodes, bX.shape[1]))
    MeanPseudotime = np.zeros(bnNodes)
    # for each branch node
    for j in range(bnNodes):
        # index associated data
        data_idx = clusters == j

        # generate node
        PseudotimeNodePositions[j] = bX[data_idx].mean(axis=0)
        MeanPseudotime[j] = bpseudotime[data_idx].mean()
    return PseudotimeNodePositions, MeanPseudotime


def interp_bins(nodes):
    """interpolate nodes evenly"""
    # euclidean distances between consecutive points
    dist = np.sqrt((np.diff(nodes, axis=0) ** 2).sum(axis=1))
    u = np.cumsum(dist)
    u = np.hstack([[0], u])
    t = np.linspace(0, u.max(), len(nodes))
    return np.array([np.interp(t, u, coord) for coord in nodes.T]).T


@nb.njit(cache=True)
def nNodes_pseudotime_nb(bX, bpseudotime, bnNodes):
    """
    Argsorts pseudotime and create nNodes bins
    with even number of points  but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    PseudotimeNodePositions = np.zeros((bnNodes, bX.shape[1]))
    MeanPseudotime = np.zeros(bnNodes)
    idx_data = np.linspace(0, len(bX), bnNodes + 1).astype(np.int64)
    for i in range(bnNodes):
        MeanPseudotime[i] = bpseudotime[
            argsort_pseudotime[idx_data[i] : idx_data[i + 1]]
        ].mean()
        for j in range(bX.shape[1]):
            PseudotimeNodePositions[i, j] = bX[
                argsort_pseudotime[idx_data[i] : idx_data[i + 1]], j
            ].mean()

    return PseudotimeNodePositions, MeanPseudotime


@nb.njit(cache=True)
def chunk(bX, bX_chunks, bps_chunks, bnNodes):
    PseudotimeNodePositions = np.full((bnNodes, bX.shape[1]), np.nan)
    MeanPseudotime = np.full(bnNodes, np.nan)
    invalidNodes = []
    for i, bX_chunk, bs_chunk in zip(np.arange(bnNodes), bX_chunks, bps_chunks):
        if len(bs_chunk):
            MeanPseudotime[i] = bs_chunk.mean()
            for j in range(bX.shape[1]):
                PseudotimeNodePositions[i, j] = bX_chunk[:, j].mean()
        else:
            invalidNodes.append(i)

    # handle edge case: empty bins node positions set to average of adjacent bins
    for i in invalidNodes:
        b, a = np.max(np.array([0, i - 1])), np.min(np.array([bnNodes - 1, i + 1]))

        MeanPseudotime[i] = np.nanmean(MeanPseudotime[np.array([b, a])])

        for j in range(bX.shape[1]):
            PseudotimeNodePositions[i, j] = np.nanmean(
                PseudotimeNodePositions[np.array([b, a]), j]
            )

    return PseudotimeNodePositions, MeanPseudotime


@nb.njit(cache=True)
def chunkv3(bX, bX_chunks, bps_chunks, bnNodes):
    PseudotimeNodePositions = np.full((bnNodes, bX.shape[1]), np.nan)
    invalidNodes = []
    for i, bX_chunk, bs_chunk in zip(np.arange(bnNodes), bX_chunks, bps_chunks):
        if len(bs_chunk):
            for j in range(bX.shape[1]):
                PseudotimeNodePositions[i, j] = bX_chunk[:, j].mean()
        else:
            invalidNodes.append(i)

    # handle edge case: empty bins node positions set to average of adjacent bins
    for i in invalidNodes:
        b, a = np.max(np.array([0, i - 1])), np.min(np.array([bnNodes - 1, i + 1]))
        if b in invalidNodes and a in invalidNodes:
            return PseudotimeNodePositions, True

        for j in range(bX.shape[1]):
            PseudotimeNodePositions[i, j] = np.nanmean(
                PseudotimeNodePositions[np.array([b, a]), j]
            )
    return PseudotimeNodePositions, False


@nb.njit(cache=True)
def nNodes_pseudotime_even_nb(bX, bpseudotime, bnNodes):
    """
    Argsorts pseudotime and create nNodes bins
    with roughly even amount of weights but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    sorted_weights = np.ones(len(bpseudotime))
    cum_arr = sorted_weights.cumsum() / sorted_weights.sum()
    idx = np.searchsorted(cum_arr, np.linspace(0, 1, bnNodes + 1)[1:-1], side="right")
    bX_chunks = nb.typed.List(np.split(bX[argsort_pseudotime], idx))
    bps_chunks = nb.typed.List(np.split(bpseudotime[argsort_pseudotime], idx))
    PseudotimeNodePositions, MeanPseudotime = chunk(bX, bX_chunks, bps_chunks, bnNodes)
    return PseudotimeNodePositions, MeanPseudotime, bX_chunks


@nb.njit(cache=True)
def nNodes_pseudotime_weighted_nb(bX, bpseudotime, bnNodes, bweights):
    """
    Argsorts pseudotime and create nNodes bins
    with roughly even amount of weights but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    sorted_weights = bweights[argsort_pseudotime]
    cum_arr = sorted_weights.cumsum() / sorted_weights.sum()
    idx = np.searchsorted(cum_arr, np.linspace(0, 1, bnNodes + 1)[1:-1], side="right")
    bX_chunks = nb.typed.List(np.split(bX[argsort_pseudotime], idx))
    bps_chunks = nb.typed.List(np.split(bpseudotime[argsort_pseudotime], idx))
    PseudotimeNodePositions, MeanPseudotime = chunk(bX, bX_chunks, bps_chunks, bnNodes)
    return PseudotimeNodePositions, MeanPseudotime, bX_chunks


@nb.njit(cache=True)
def nNodes_pseudotime_even_nbv3(bX, bpseudotime, bnNodes):
    """
    Argsorts pseudotime and create nNodes bins
    with roughly even amount of weights but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    sorted_weights = np.ones(len(bpseudotime))
    cum_arr = sorted_weights.cumsum() / sorted_weights.sum()
    idx = np.searchsorted(cum_arr, np.linspace(0, 1, bnNodes + 1)[1:-1], side="right")
    bX_chunks = nb.typed.List(np.split(bX[argsort_pseudotime], idx))
    bps_chunks = nb.typed.List(np.split(bpseudotime[argsort_pseudotime], idx))
    PseudotimeNodePositions = chunkv3(bX, bX_chunks, bps_chunks, bnNodes)[0]
    return PseudotimeNodePositions, bX_chunks


@nb.njit(cache=True)
def nNodes_pseudotime_weighted_nbv3(bX, bpseudotime, bnNodes, bweights):
    """
    Argsorts pseudotime and create nNodes bins
    with roughly even amount of weights but potentially uneven amount of pseudotime
    """
    argsort_pseudotime = np.argsort(bpseudotime)
    sorted_weights = bweights[argsort_pseudotime]
    cum_arr = sorted_weights.cumsum() / sorted_weights.sum()
    idx = np.searchsorted(cum_arr, np.linspace(0, 1, bnNodes + 1)[1:-1], side="right")
    bX_chunks = nb.typed.List(np.split(bX[argsort_pseudotime], idx))
    bps_chunks = nb.typed.List(np.split(bpseudotime[argsort_pseudotime], idx))
    PseudotimeNodePositions, invalid = chunkv3(bX, bX_chunks, bps_chunks, bnNodes)
    return PseudotimeNodePositions, bX_chunks, invalid


@nb.njit
def bin_points(bpseudotime, bnNodes):
    bins = np.linspace(bpseudotime.min(), bpseudotime.max(), bnNodes + 1)
    bins[0] = -np.inf
    clusters = np.zeros(len(bpseudotime))
    count = np.zeros(bnNodes)
    for i in range(bnNodes):
        idx_bin = (bpseudotime > bins[i]) & (bpseudotime <= bins[i + 1])
        clusters[idx_bin] = i
        count[i] = idx_bin.sum()
    return count, bins, clusters


@nb.njit(cache=True)
def bin_pseudotime_nb(bX, bpseudotime, bnNodes):
    """
    Argsorts pseudotime and create nNodes bins
    with even amounts of pseudotime but potentially uneven number of points
    """
    # ---create nodes with uniformly spread pseudotime
    count, bins, clusters = bin_points(bpseudotime, bnNodes)

    validNodes = count > 0

    PseudotimeNodePositions = np.zeros((bnNodes, bX.shape[1]))
    MeanPseudotime = np.zeros(bnNodes)
    # for each branch node
    for i in np.where(validNodes)[0]:

        # index associated data
        data_idx = clusters == i

        # generate node
        MeanPseudotime[i] = bpseudotime[data_idx].mean()

        for j in range(bX.shape[1]):
            PseudotimeNodePositions[i, j] = bX[data_idx, j].mean()

    # handle edge case: empty bins node positions set to average of adjacent bins
    for i in np.where(~validNodes)[0]:
        b, a = np.max(np.array([0, i - 1])), np.min(np.array([bnNodes - 1, i + 1]))

        MeanPseudotime[i] = MeanPseudotime[np.array([b, a])].mean()

        for j in range(bX.shape[1]):
            PseudotimeNodePositions[i, j] = PseudotimeNodePositions[
                np.array([b, a]), j
            ].mean()

    return PseudotimeNodePositions, MeanPseudotime


@nb.njit(cache=True)
def interp_bins_nb(nodes):
    """interpolate nodes evenly"""
    # euclidean distances between consecutive points
    dist = np.sqrt(((nodes[1:] - nodes[:-1]) ** 2).sum(axis=1))
    u = np.cumsum(dist)
    u = np.concatenate((np.array([0.0]), u))

    t = np.linspace(0, u.max(), len(nodes))

    out = np.zeros(nodes.shape)
    for i in range(nodes.shape[1]):
        out[:, i] = np.interp(t, u, nodes[:, i])

    return out


def gen_pseudotime_centroids(
    X, pseudotime, branches_single_end, branches_dataidx, PointWeights
):
    """generate pseudotime centroids for each branch of the graph"""

    # ignore branches with associated ndatapoints less than nNodes
    nonempty_branches_dataidx = {}
    nonempty_branches_single_end = {}
    nonempty_nNodes = 0
    for (k, bdata), bnodes in zip(
        branches_dataidx.items(), branches_single_end.values()
    ):
        if len(np.where(bdata)[0]) >= len(bnodes):
            nonempty_branches_dataidx[k] = bdata
            nonempty_branches_single_end[k] = bnodes
            nonempty_nNodes += len(branches_single_end[k])

    PseudotimeNodePositions = np.zeros((nonempty_nNodes, X.shape[1]))
    MeanPseudotime = np.zeros(nonempty_nNodes)

    n = 0
    for (
        k,
        bdata,
    ) in nonempty_branches_dataidx.items():  # for data associated with each branch

        # branch data points, pseudotime
        bX, bpseudotime, bweights = X[bdata], pseudotime[bdata], PointWeights[bdata]
        bNodes = nonempty_branches_single_end[k]
        bnNodes = len(branches_single_end[k])

        # generate paths ps node positions
        _PseudotimeNodePositions, _MeanPseudotime = nNodes_pseudotime_weighted_nb(
            bX, bpseudotime, bnNodes, bweights
        )
        _PseudotimeNodePositions = interp_bins_nb(_PseudotimeNodePositions)

        # add them to common array
        PseudotimeNodePositions[bNodes] += _PseudotimeNodePositions
        MeanPseudotime[bNodes] += _MeanPseudotime

    return PseudotimeNodePositions, MeanPseudotime, nonempty_branches_single_end


def old_gen_pseudotime_centroids_by_path(
    X, pseudotime, paths, paths_dataidx, PointWeights
):
    """generate pseudotime centroids for each path of the graph"""

    # ---ignore paths with associated ndatapoints less than nNodes
    nonempty_paths_dataidx = {}
    nonempty_paths = {}
    for (k, bdata), bnodes in zip(paths_dataidx.items(), paths.values()):
        if len(np.where(bdata)[0]) >= len(bnodes):
            nonempty_paths_dataidx[k] = bdata
            nonempty_paths[k] = bnodes

    # ---generate pseudotime nodes
    nodes_list = list(itertools.chain.from_iterable(nonempty_paths.values()))

    nonempty_nNodes = len(set(nodes_list))
    nNodes_bincount = np.bincount(nodes_list)

    PseudotimeNodePositions = np.zeros((nonempty_nNodes, X.shape[1]))
    MeanPseudotime = np.zeros(nonempty_nNodes)

    for (
        k,
        bdata,
    ) in nonempty_paths_dataidx.items():  # for data associated with each branch

        # paths data points, pseudotime, Nodes, nNodes
        bX, bpseudotime, bweights = X[bdata], pseudotime[bdata], PointWeights[bdata]
        bNodes = nonempty_paths[k]
        bnNodes = len(bNodes)

        # generate paths ps node positions
        (
            _PseudotimeNodePositions,
            _MeanPseudotime,
            bX_chunks,
        ) = nNodes_pseudotime_weighted_nb(bX, bpseudotime, bnNodes, bweights)
        if np.any(np.array([len(b) for b in bX_chunks]) < 2):
            # generate paths ps node positions
            (
                _PseudotimeNodePositions,
                _MeanPseudotime,
                bX_chunks,
            ) = nNodes_pseudotime_even_nb(
                bX,
                bpseudotime,
                bnNodes,
            )

        _PseudotimeNodePositions = np.vstack(
            [
                bX_chunks[i][vq(_PseudotimeNodePositions[[i]], bX_chunks[i])[0][0]]
                for i in range(bnNodes)
            ]
        )
        _PseudotimeNodePositions = interp_bins_nb(_PseudotimeNodePositions)

        # add them to common array
        PseudotimeNodePositions[bNodes] += _PseudotimeNodePositions
        MeanPseudotime[bNodes] += _MeanPseudotime

    # divide to average nodes that appear in multiple paths
    PseudotimeNodePositions /= nNodes_bincount[:, None]
    MeanPseudotime /= nNodes_bincount
    return PseudotimeNodePositions, MeanPseudotime, nonempty_paths


def old_gen_pseudotime_centroids_by_path(
    X, NodePositions, pseudotime, paths, paths_dataidx, PointWeights
):
    """generate pseudotime centroids for each path of the graph"""

    # ---generate pseudotime nodes
    nodes_list = list(itertools.chain.from_iterable(paths.values()))

    nNodes = len(NodePositions)
    nNodes_bincount = np.bincount(nodes_list)

    PseudotimeNodePositions = np.zeros((nNodes, X.shape[1]))
    MeanPseudotime = np.zeros(nNodes)

    for (
        k,
        bdata,
    ) in paths_dataidx.items():  # for data associated with each branch

        # paths data points, pseudotime, Nodes, nNodes
        bX, bpseudotime, bweights = X[bdata], pseudotime[bdata], PointWeights[bdata]
        bNodes = paths[k]
        bnNodes = len(bNodes)

        if len(np.where(bdata)[0]) <= bnNodes:
            # ignore paths with associated ndatapoints less than nNodes
            PseudotimeNodePositions[bNodes] += NodePositions[bNodes]
        else:
            # generate paths ps node positions
            (
                _PseudotimeNodePositions,
                _MeanPseudotime,
                bX_chunks,
            ) = nNodes_pseudotime_weighted_nb(bX, bpseudotime, bnNodes, bweights)
            if np.any(np.array([len(b) for b in bX_chunks]) < 2):
                # generate paths ps node positions
                (
                    _PseudotimeNodePositions,
                    _MeanPseudotime,
                    bX_chunks,
                ) = nNodes_pseudotime_even_nb(
                    bX,
                    bpseudotime,
                    bnNodes,
                )

            _PseudotimeNodePositions = np.vstack(
                [
                    bX_chunks[i][vq(_PseudotimeNodePositions[[i]], bX_chunks[i])[0][0]]
                    for i in range(bnNodes)
                ]
            )
            _PseudotimeNodePositions = interp_bins_nb(_PseudotimeNodePositions)

            # add them to common array
            PseudotimeNodePositions[bNodes] += _PseudotimeNodePositions
            MeanPseudotime[bNodes] += _MeanPseudotime

    # divide to average nodes that appear in multiple paths
    PseudotimeNodePositions /= nNodes_bincount[:, None]
    MeanPseudotime /= nNodes_bincount
    return PseudotimeNodePositions, MeanPseudotime, paths


def _gen_pseudotime_centroids_singlepath(bX, bpseudotime, bnNodes, bweights):
    # generate paths ps node positions
    (_PseudotimeNodePositions, bX_chunks, invalid) = nNodes_pseudotime_weighted_nbv3(
        bX, bpseudotime, bnNodes, bweights
    )
    if invalid:
        # generate paths ps node positions
        (_PseudotimeNodePositions, bX_chunks,) = nNodes_pseudotime_even_nbv3(
            bX,
            bpseudotime,
            bnNodes,
        )

    tmp = []
    for i in range(bnNodes):
        if len(bX_chunks[i]) == 0:
            tmp.append(_PseudotimeNodePositions[i])
        else:
            tmp.append(
                bX_chunks[i][vq(_PseudotimeNodePositions[[i]], bX_chunks[i])[0][0]]
            )
    _PseudotimeNodePositions = np.vstack(tmp)
    _PseudotimeNodePositions = interp_bins_nb(_PseudotimeNodePositions)
    return _PseudotimeNodePositions


def gen_pseudotime_centroids_by_path(
    X,
    NodePositions,
    pseudotime,
    paths,
    paths_dataidx,
    PointWeights,
    cache_PseudotimeNodePositions={},
):
    """generate pseudotime centroids for each path of the graph"""

    # ---generate pseudotime nodes
    nodes_list = list(itertools.chain.from_iterable(paths.values()))
    nNodes = len(NodePositions)
    nNodes_bincount = np.bincount(nodes_list)
    PseudotimeNodePositions = np.zeros((nNodes, X.shape[1]))

    for (
        k,
        bdata,
    ) in paths_dataidx.items():  # for data associated with each branch

        # paths data points, pseudotime, Nodes, nNodes
        bX, bpseudotime, bweights = X[bdata], pseudotime[bdata], PointWeights[bdata]
        bNodes = paths[k]
        bnNodes = len(bNodes)

        if len(np.where(bdata)[0]) <= bnNodes:
            # ignore paths with associated ndatapoints less than nNodes
            PseudotimeNodePositions[bNodes] += NodePositions[bNodes]
        else:
            bytes_bdata = bdata.tobytes()
            if (bytes_bdata, bnNodes) in cache_PseudotimeNodePositions.keys():
                # retrieve precomputed path
                _PseudotimeNodePositions = cache_PseudotimeNodePositions[
                    (bytes_bdata, bnNodes)
                ]
            else:
                # unseen params, generate the path
                _PseudotimeNodePositions = _gen_pseudotime_centroids_singlepath(
                    bX, bpseudotime, bnNodes, bweights
                )
                cache_PseudotimeNodePositions[
                    (bytes_bdata, bnNodes)
                ] = _PseudotimeNodePositions

            # add them to common array
            PseudotimeNodePositions[bNodes] += _PseudotimeNodePositions

    # divide to average nodes that appear in multiple paths
    PseudotimeNodePositions /= nNodes_bincount[:, None]
    return PseudotimeNodePositions, paths, cache_PseudotimeNodePositions


# ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes and merge pseudotime and elpigraph nodesp, elasticmatrix
def augment_graph(
    NodePositions,
    Edges,
    PseudotimeNodePositions,
    branches_single_end,
    Mus,
    Lambdas,
    LinkMu,
    LinkLambda,
):
    """
    generate a graph merging node positions and pseudotime node positions
    with one edge between each of their nodes.
    pseudotime node positions and edges are placed as the top rows of the matrices
    """
    # ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes
    # ordering of nodespositions in the graph (corresponding to pseudotime nodes)
    NodesOrder = np.array([n for b in branches_single_end.values() for n in b])
    PseudotimeNodes = np.arange(len(PseudotimeNodePositions))
    # link pseudotime nodes and graph nodes
    LinkEdges = np.array(
        list(zip(PseudotimeNodes, NodesOrder + len(PseudotimeNodePositions)))
    )
    LinkMus = np.repeat(LinkMu, len(PseudotimeNodePositions))
    LinkLambdas = np.repeat(LinkLambda, len(PseudotimeNodePositions))

    # -----merge pseudotime and graph nodepositions, elasticmatrix
    MergedNodePositions = np.concatenate((PseudotimeNodePositions, NodePositions))
    MergedEdges = np.concatenate((LinkEdges, Edges + len(PseudotimeNodePositions)))
    MergedLambdas = np.concatenate((LinkLambdas, Lambdas))
    MergedMus = np.concatenate((LinkMus, Mus))
    MergedElasticMatrix = Encode2ElasticMatrix(MergedEdges, MergedLambdas, MergedMus)

    return (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        MergedLambdas,
        MergedMus,
    )


def augment_graph_by_path(
    NodePositions,
    Edges,
    PseudotimeNodePositions,
    Mus,
    Lambdas,
    LinkMu,
    LinkLambda,
):
    """
    generate a graph merging node positions and pseudotime node positions
    with one edge between each of their nodes.
    pseudotime node positions and edges are placed as the top rows of the matrices
    """
    # ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes
    # ordering of nodespositions in the graph (corresponding to pseudotime nodes)
    nPsNodes = len(PseudotimeNodePositions)
    # link pseudotime nodes and graph nodes
    LinkEdges = np.array(list(zip(np.arange(nPsNodes), np.arange(nPsNodes) + nPsNodes)))
    LinkMus = np.repeat(LinkMu, nPsNodes)
    LinkLambdas = np.repeat(LinkLambda, nPsNodes)

    # -----merge pseudotime and graph nodepositions, elasticmatrix
    MergedNodePositions = np.concatenate((PseudotimeNodePositions, NodePositions))
    MergedEdges = np.concatenate((LinkEdges, Edges + nPsNodes))
    MergedLambdas = np.concatenate((LinkLambdas, Lambdas))
    MergedMus = np.concatenate((LinkMus, Mus))
    MergedElasticMatrix = Encode2ElasticMatrix(MergedEdges, MergedLambdas, MergedMus)

    return (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        MergedLambdas,
        MergedMus,
    )


# ---main functions
def gen_pseudotime_augmented_graph(
    X,
    SquaredX,
    NodePositions,
    ElasticMatrix,
    pseudotime,
    root_node,
    LinkMu,
    LinkLambda,
    PointWeights,
    TrimmingRadius=float("inf"),
):
    # ------extract oriented branches and associated data
    Edges, Lambdas, Mus = DecodeElasticMatrix(ElasticMatrix)

    # handle edge case: circle or curve topology
    if np.bincount(Edges.flat).max() == 2:
        print("WARNING: fitting circle or curve topology")
        branches_single_end = get_circle_or_curve(Edges, root_node)
    else:
        tree, branches, branches_single_end = get_tree(Edges, root_node)
    branches_dataidx = partition_data_by_branch(
        X, SquaredX, NodePositions, branches_single_end, TrimmingRadius
    )

    # ------generate pseudotime centroid branches
    (
        PseudotimeNodePositions,
        MeanPseudotime,
        nonempty_branches_single_end,
    ) = gen_pseudotime_centroids(
        X, pseudotime, branches_single_end, branches_dataidx, PointWeights
    )

    # ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes and merge pseudotime and elpigraph nodesp, elasticmatrix
    (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        MergedLambdas,
        MergedMus,
    ) = augment_graph(
        NodePositions,
        Edges,
        PseudotimeNodePositions,
        nonempty_branches_single_end,
        Mus,
        Lambdas,
        LinkMu,
        LinkLambda,
    )
    return (
        MeanPseudotime,
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        len(PseudotimeNodePositions),
    )


def old_gen_pseudotime_augmented_graph_by_path(
    X,
    SquaredX,
    NodePositions,
    ElasticMatrix,
    pseudotime,
    root_node,
    LinkMu,
    LinkLambda,
    PointWeights,
    TrimmingRadius=float("inf"),
    partition=None,
):
    # ------extract oriented paths and associated data
    Edges, Lambdas, Mus = DecodeElasticMatrix(ElasticMatrix)

    paths = get_shortest_paths(Edges, root_node)
    if partition is None:
        paths_dataidx = partition_data_by_branch(
            X, SquaredX, NodePositions, paths, TrimmingRadius
        )
    else:
        paths_dataidx = partition_data_by_branch_precomp(partition, paths)

    # ------generate pseudotime centroid branches
    (
        PseudotimeNodePositions,
        MeanPseudotime,
        nonempty_paths,
    ) = old_gen_pseudotime_centroids_by_path(
        X, NodePositions, pseudotime, paths, paths_dataidx, PointWeights
    )

    # ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes and merge pseudotime and elpigraph nodesp, elasticmatrix
    (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        MergedLambdas,
        MergedMus,
    ) = augment_graph_by_path(
        NodePositions,
        Edges,
        PseudotimeNodePositions,
        Mus,
        Lambdas,
        LinkMu,
        LinkLambda,
    )
    return (
        MeanPseudotime,
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        len(PseudotimeNodePositions),
    )


def gen_pseudotime_augmented_graph_by_path(
    X,
    SquaredX,
    NodePositions,
    ElasticMatrix,
    pseudotime,
    root_node,
    LinkMu,
    LinkLambda,
    PointWeights,
    TrimmingRadius=float("inf"),
    partition=None,
    cache_PseudotimeNodePositions={},
):
    # ------extract oriented paths and associated data
    Edges, Lambdas, Mus = DecodeElasticMatrix(ElasticMatrix)

    paths = get_shortest_paths(Edges, root_node)
    if partition is None:
        paths_dataidx = partition_data_by_branch(
            X, SquaredX, NodePositions, paths, TrimmingRadius
        )
    else:
        paths_dataidx = partition_data_by_branch_precomp(partition, paths)

    # ------generate pseudotime centroid branches
    (
        PseudotimeNodePositions,
        nonempty_paths,
        cache_PseudotimeNodePositions,
    ) = gen_pseudotime_centroids_by_path(
        X,
        NodePositions,
        pseudotime,
        paths,
        paths_dataidx,
        PointWeights,
        cache_PseudotimeNodePositions,
    )

    # ------for each branch, create elastic edges between pseudotime nodes & elpigraph nodes and merge pseudotime and elpigraph nodesp, elasticmatrix
    (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        MergedLambdas,
        MergedMus,
    ) = augment_graph_by_path(
        NodePositions,
        Edges,
        PseudotimeNodePositions,
        Mus,
        Lambdas,
        LinkMu,
        LinkLambda,
    )
    return (
        MergedNodePositions,
        MergedElasticMatrix,
        MergedEdges,
        len(PseudotimeNodePositions),
        cache_PseudotimeNodePositions,
    )




from elpigraph import __version__
from distutils.version import LooseVersion
import sphinx
import sys
import os
import sphinx_rtd_theme

sys.path.append("../elpigraph/")

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "numpydoc",
]

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = False
napoleon_custom_sections = [("Params", "Parameters")]

# this is needed for some reason...
# see https://github.com/numpy/numpydoc/issues/69
numpydoc_show_class_members = False

# pngmath / imgmath compatibility layer for different sphinx versions
if LooseVersion(sphinx.__version__) < LooseVersion("1.4"):
    extensions.append("sphinx.ext.pngmath")
else:
    extensions.append("sphinx.ext.imgmath")

autodoc_default_flags = ["members", "inherited-members"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# generate autosummary even if no references
autosummary_generate = True

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# Generate the plots for the gallery
# plot_gallery = True

# The master toctree document.
master_doc = "index"

project = "elpigraph-python"
copyright = "2020, Jonathan Bac"

# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "_templates"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Custom style
# html_style = "css/project-template.css"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "elpigraph-doc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "letterpaper",
    # The font size ('10pt', '11pt' or '12pt').
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    "preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        "elpigraph-python.tex",
        "elpigraph Documentation",
        "Jonathan Bac",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#    (
#        "index",
#        "elpigraph-python",
#        u"elpigraph-python Documentation",
#        [u"Jonathan Bac"],
#        1,
#    )
# ]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = [
#    (
#        "index",
#        "elpigraph-python",
#        u"elpigraph-python Documentation",
#        u"Jonathan Bac",
#        "elpigraph-python",
#        "One line description of project.",
#        "Miscellaneous",
#    ),
# ]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False


# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx configuration
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/{.major}".format(sys.version_info),
        None,
    ),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "sklearn": ("http://scikit-learn.org/stable", None),
}

# sphinx-gallery configuration
# sphinx_gallery_conf = {
#    'doc_module': 'skdim',
#    'backreferences_dir': os.path.join('generated'),
#    'reference_url': {
#        'skdim': None}
# }


def setup(app):
    # a copy button to copy snippet of code from the documentation
    app.add_stylesheet("css/custom.css")


# -- Options for other output ------------------------------------------
import inspect
from pathlib import Path
from typing import Optional, Union, Mapping
import logging
from sphinx.application import Sphinx
from sphinx.ext import autosummary

logger = logging.getLogger(__name__)
author = "Jonathan Bac"
title = "elpigraph-python"
title_doc = f"{project} documentation"

latex_documents = [(master_doc, f"{project}.tex", title_doc, author, "manual")]
man_pages = [(master_doc, project, title_doc, [author], 1)]
texinfo_documents = [
    (master_doc, project, title_doc, author, project, title, "Miscellaneous")
]


# -- generate_options override ------------------------------------------


def process_generate_options(app: Sphinx):
    genfiles = app.config.autosummary_generate

    if genfiles and not hasattr(genfiles, "__len__"):
        env = app.builder.env
        genfiles = [
            env.doc2path(x, base=None)
            for x in env.found_docs
            if Path(env.doc2path(x)).is_file()
        ]
    if not genfiles:
        return

    from sphinx.ext.autosummary.generate import generate_autosummary_docs

    ext = app.config.source_suffix
    genfiles = [
        genfile + (not genfile.endswith(tuple(ext)) and ext[0] or "")
        for genfile in genfiles
    ]

    suffix = autosummary.get_rst_suffix(app)
    if suffix is None:
        return

    generate_autosummary_docs(
        genfiles,
        builder=app.builder,
        warn=logger.warning,
        info=logger.info,
        suffix=suffix,
        base_path=app.srcdir,
        imported_members=True,
        app=app,
    )


autosummary.process_generate_options = process_generate_options


# -- GitHub URLs for class and method pages ------------------------------------------


def get_obj_module(qualname):
    """Get a module/class/attribute and its original module by qualname"""
    modname = qualname
    classname = None
    attrname = None
    while modname not in sys.modules:
        attrname = classname
        modname, classname = modname.rsplit(".", 1)

    # retrieve object and find original module name
    if classname:
        cls = getattr(sys.modules[modname], classname)
        modname = cls.__module__
        obj = getattr(cls, attrname) if attrname else cls
    else:
        obj = None

    return obj, sys.modules[modname]


def get_linenos(obj):
    """Get an object’s line numbers"""
    try:
        lines, start = inspect.getsourcelines(obj)
    except TypeError:
        return None, None
    else:
        return start, start + len(lines) - 1


# set project_dir: project/docs/source/conf.py/../../.. → project/
project_dir = Path(__file__).parent.parent.parent
github_url_scvelo = "https://github.com/j-bac/elpigraph-python/tree/master"
from pathlib import PurePosixPath


def modurl(qualname):
    """Get the full GitHub URL for some object’s qualname."""
    obj, module = get_obj_module(qualname)
    github_url = github_url_scvelo
    # try:
    path = PurePosixPath(
        Path(module.__file__).resolve().relative_to(project_dir)
    )
    # except ValueError:
    # trying to document something from another package
    # github_url = (
    #    if "read" in qualname
    # )
    # path = "/".join(module.__file__.split("/")[-2:])
    start, end = get_linenos(obj)
    fragment = f"#L{start}-L{end}" if start and end else ""
    return f"{github_url}/{path}{fragment}"


def api_image(qualname: str) -> Optional[str]:
    path = Path(__file__).parent / f"{qualname}.png"
    print(path, path.is_file())
    return (
        f".. image:: {path.name}\n   :width: 200\n   :align: right"
        if path.is_file()
        else ""
    )


# modify the default filters
from jinja2.defaults import DEFAULT_FILTERS

DEFAULT_FILTERS.update(modurl=modurl, api_image=api_image)

# -- Override some classnames in autodoc --------------------------------------------

import sphinx_autodoc_typehints

qualname_overrides = {
    "anndata.base.AnnData": "anndata.AnnData",
}

fa_orig = sphinx_autodoc_typehints.format_annotation


def format_annotation(annotation):
    if getattr(annotation, "__origin__", None) is Union or hasattr(
        annotation, "__union_params__"
    ):
        params = getattr(annotation, "__union_params__", None) or getattr(
            annotation, "__args__", None
        )
        return ", ".join(map(format_annotation, params))
    if getattr(annotation, "__origin__", None) is Mapping:
        return ":class:`~typing.Mapping`"
    if inspect.isclass(annotation):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = qualname_overrides.get(full_name)
        if override is not None:
            return f":py:class:`~{qualname_overrides[full_name]}`"
    return fa_orig(annotation)


sphinx_autodoc_typehints.format_annotation = format_annotation


# -- Prettier Param docs --------------------------------------------

from typing import Dict, List, Tuple
from docutils import nodes
from sphinx import addnodes
from sphinx.domains.python import PyTypedField, PyObject
from sphinx.environment import BuildEnvironment


class PrettyTypedField(PyTypedField):
    list_type = nodes.definition_list

    def make_field(
        self,
        types: Dict[str, List[nodes.Node]],
        domain: str,
        items: Tuple[str, List[nodes.inline]],
        env: BuildEnvironment = None,
    ) -> nodes.field:
        def makerefs(rolename, name, node):
            return self.make_xrefs(rolename, domain, name, node, env=env)

        def handle_item(
            fieldarg: str, content: List[nodes.inline]
        ) -> nodes.definition_list_item:
            head = nodes.term()
            head += makerefs(self.rolename, fieldarg, addnodes.literal_strong)
            fieldtype = types.pop(fieldarg, None)
            if fieldtype is not None:
                head += nodes.Text(" : ")
                if len(fieldtype) == 1 and isinstance(
                    fieldtype[0], nodes.Text
                ):
                    (text_node,) = fieldtype  # type: nodes.Text
                    head += makerefs(
                        self.typerolename,
                        text_node.astext(),
                        addnodes.literal_emphasis,
                    )
                else:
                    head += fieldtype

            body_content = nodes.paragraph("", "", *content)
            body = nodes.definition("", body_content)

            return nodes.definition_list_item("", head, body)

        fieldname = nodes.field_name("", self.label)
        if len(items) == 1 and self.can_collapse:
            fieldarg, content = items[0]
            bodynode = handle_item(fieldarg, content)
        else:
            bodynode = self.list_type()
            for fieldarg, content in items:
                bodynode += handle_item(fieldarg, content)
        fieldbody = nodes.field_body("", bodynode)
        return nodes.field("", fieldname, fieldbody)


# replace matching field types with ours
PyObject.doc_field_types = [
    PrettyTypedField(
        ft.name,
        names=ft.names,
        typenames=ft.typenames,
        label=ft.label,
        rolename=ft.rolename,
        typerolename=ft.typerolename,
        can_collapse=ft.can_collapse,
    )
    if isinstance(ft, PyTypedField)
    else ft
    for ft in PyObject.doc_field_types
]



import numpy as np
import igraph
import scipy.optimize
import copy
import warnings
from .src.graphs import ConstructGraph, GetSubGraph, GetBranches
from .src.core import PartitionData
from .src.distutils import PartialDistance
from .src.reporting import project_point_onto_graph, project_point_onto_edge


def ExtendLeaves(
    X,
    PG,
    Mode="QuantDists",
    ControlPar=0.9,
    DoSA=True,
    DoSA_maxiter=2000,
    LeafIDs=None,
    TrimmingRadius=float("inf"),
    # PlotSelected=False,
):
    """
    Extend leaves with additional nodes

    X numeric matrix, the data matrix
    TargetPG list, the ElPiGraph structure to extend
    LeafIDs integer vector, the id of nodes to extend. If NULL, all the vertices will be extended.
    TrimmingRadius positive numeric, the trimming radius used to control distance
    ControlPar positive numeric, the paramter used to control the contribution of the different data points
    DoSA bollean, should optimization (via simulated annealing) be performed when Mode = "QuantDists"?
    Mode string, the mode used to extend the graph. "QuantCentroid" and "WeigthedCentroid" are currently implemented
    PlotSelected boolean, should a diagnostic plot be visualized

    Return
    -------
    The extended ElPiGraph structure

    The value of ControlPar has a different interpretation depending on the valus of Mode. In each case, for only the extreme points,
    i.e., the points associated with the leaf node that do not have a projection on any edge are considered.

    If Mode = "QuantCentroid", for each leaf node, the extreme points are ordered by their distance from the node
    and the centroid of the points farther away than the ControlPar is returned.

    If Mode = "WeightedCentroid", for each leaf node, a weight is computed for each points by raising the distance to the ControlPar power.
    Hence, larger values of ControlPar result in a larger influence of points farther from the node

    If Mode = "QuantDists", for each leaf node, ... will write it later
    """

    TargetPG = copy.deepcopy(PG)
    # Generate net
    Net = ConstructGraph(PrintGraph=TargetPG)

    # get leafs
    if LeafIDs is None:
        LeafIDs = np.where(np.array(Net.degree()) == 1)[0]

    # check LeafIDs
    if np.any(np.array(Net.degree(LeafIDs)) > 1):
        raise ValueError("Only leaf nodes can be extended")

    # and their neigh
    Nei = Net.neighborhood(LeafIDs, order=1)

    # and put stuff together
    NeiVect = list(
        map(lambda x: set(Nei[x]).difference([LeafIDs[x]]), range(len(Nei)))
    )
    NeiVect = np.array([j for i in NeiVect for j in list(i)])
    NodesMat = np.hstack((LeafIDs[:, None], NeiVect[:, None]))

    # project data on the nodes
    PD = PartitionData(
        X=X,
        NodePositions=TargetPG["NodePositions"],
        MaxBlockSize=10000000,
        TrimmingRadius=TrimmingRadius,
        SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
    )

    # Keep track of the new nodes IDs
    NodeID = len(TargetPG["NodePositions"]) - 1

    init = False
    NNPos = None
    NEdgs = None
    UsedNodes = []

    # for each leaf
    for i in range(len(NodesMat)):

        if np.sum(PD[0] == NodesMat[i, 0]) == 0:
            continue

        # generate the new node id
        NodeID = NodeID + 1

        # get all the data associated with the leaf node
        tData = X[(PD[0] == NodesMat[i, 0]).flatten(), :]

        # and project them on the edge
        Proj = project_point_onto_edge(
            X=X[(PD[0] == NodesMat[i, 0]).flatten(), :],
            NodePositions=TargetPG["NodePositions"],
            Edge=NodesMat[i, :],
        )

        # Select the distances of the associated points
        Dists = PD[1][PD[0] == NodesMat[i, 0]]

        # Set distances of points projected on beyond the initial position of the edge to 0
        Dists[Proj["Projection_Value"] >= 0] = 0

        if Mode == "QuantCentroid":
            ThrDist = np.quantile(Dists[Dists > 0], ControlPar)
            SelPoints = np.where(Dists >= ThrDist)[0]

            print(
                len(SelPoints),
                " points selected to compute the centroid while extending"
                " node",
                NodesMat[i, 0],
            )

            if len(SelPoints) > 1:
                NN = np.mean(tData[SelPoints, :], axis=0, keepdims=1)

            else:
                NN = tData[SelPoints, :]

            # Initialize the new nodes and edges
            if not init:
                init = True
                NNPos = NN.copy()
                NEdgs = np.array([[NodesMat[i, 0], NodeID]])
                UsedNodes.extend(
                    np.where(PD[0].flatten() == NodesMat[i, 0])[0][SelPoints]
                )
            else:
                NNPos = np.vstack((NNPos, NN))
                NEdgs = np.vstack(
                    (NEdgs, np.array([[NodesMat[i, 0], NodeID]]))
                )
                UsedNodes.extend(
                    list(
                        np.where(PD[0].flatten() == NodesMat[i, 0])[0][
                            SelPoints
                        ]
                    )
                )

        if Mode == "WeightedCentroid":
            Dist2 = Dists ** (2 * ControlPar)
            Wei = Dist2 / np.max(Dist2)

            if len(Wei) > 1:
                NN = np.sum(tData * Wei[:, None], axis=0) / np.sum(Wei)

            else:
                NN = tData

            # Initialize the new nodes and edges
            if not init:
                init = True
                NNPos = NN.copy()
                NEdgs = np.array([NodesMat[i, 0], NodeID])
                WeiVal = list(Wei)
                UsedNodes.extend(
                    list(np.where(PD[0].flatten() == NodesMat[i, 0])[0])
                )
            else:
                NNPos = np.vstack((NNPos, NN))
                NEdgs = np.vstack((NEdgs, np.array([NodesMat[i, 0], NodeID])))

                UsedNodes.extend(
                    list(np.where(PD[0].flatten() == NodesMat[i, 0])[0])
                )
                WeiVal.extend(list(Wei))

        if Mode == "QuantDists":

            if sum(Dists > 0) == 0:
                continue

            if sum(Dists > 0) > 1 and len(tData) > 1:
                tData_Filtered = tData[Dists > 0, :]

                def DistFun(NodePosition):
                    return np.quantile(
                        project_point_onto_edge(
                            X=tData_Filtered,
                            NodePositions=np.vstack(
                                (
                                    TargetPG["NodePositions"][
                                        NodesMat[i, 0],
                                    ],
                                    NodePosition,
                                )
                            ),
                            Edge=[0, 1],
                        )["Distance_Squared"],
                        ControlPar,
                    )

                StartingPoint = tData_Filtered[
                    np.argmin(
                        np.array(
                            [
                                DistFun(tData_Filtered[i])
                                for i in range(len(tData_Filtered))
                            ]
                        )
                    ),
                    :,
                ]

                if DoSA:

                    print(
                        "Performing simulated annealing. This may take a while"
                    )
                    StartingPoint = scipy.optimize.dual_annealing(
                        DistFun,
                        bounds=list(
                            zip(
                                np.min(tData_Filtered, axis=0),
                                np.max(tData_Filtered, axis=0),
                            )
                        ),
                        x0=StartingPoint,
                        maxiter=DoSA_maxiter,
                    )["x"]

                Projections = project_point_onto_edge(
                    X=tData_Filtered,
                    NodePositions=np.vstack(
                        (
                            TargetPG["NodePositions"][NodesMat[i, 0], :],
                            StartingPoint,
                        )
                    ),
                    Edge=[0, 1],
                    ExtProj=True,
                )

                SelId = np.argmax(
                    PartialDistance(
                        Projections["X_Projected"],
                        np.array(
                            [TargetPG["NodePositions"][NodesMat[i, 0], :]]
                        ),
                    )
                )

                StartingPoint = Projections["X_Projected"][SelId, :]

            else:
                StartingPoint = tData[Dists > 0, :]

            if not init:
                init = True
                NNPos = StartingPoint[None].copy()
                NEdgs = np.array([[NodesMat[i, 0], NodeID]])
                UsedNodes.extend(
                    list(np.where(PD[0].flatten() == NodesMat[i, 0])[0])
                )
            else:
                NNPos = np.vstack((NNPos, StartingPoint[None]))
                NEdgs = np.vstack(
                    (NEdgs, np.array([[NodesMat[i, 0], NodeID]]))
                )
                UsedNodes.extend(
                    list(np.where(PD[0].flatten() == NodesMat[i, 0])[0])
                )

    # plot(X)
    # points(TargetPG$NodePositions, col="red")
    # points(NNPos, col="blue")
    #
    try:
        print(NNPos.shape)
    except:
        print("failed")
    try:
        print(TargetPG["NodePositions"].shape)
    except:
        pass
    TargetPG["NodePositions"] = np.vstack((TargetPG["NodePositions"], NNPos))
    TargetPG["Edges"] = [
        np.vstack((TargetPG["Edges"][0], NEdgs)),  # edges
        np.append(
            TargetPG["Edges"][1], np.repeat(np.nan, len(NEdgs))
        ),  # lambdas
        # np.append(TargetPG["Edges"][2], np.repeat(np.nan, len(NEdgs))),
    ]  ##### mus become lambda in R, bug ??

    #    if PlotSelected:
    #        if Mode == "QuantCentroid":
    #            Cats = ["Unused"] * len(X)
    #            if UsedNodes:
    #                Cats[UsedNodes] = "Used"
    #
    #            p = PlotPG(X=X, TargetPG=TargetPG, GroupsLab=Cats)
    #            print(p)
    #
    #        if Mode == "WeightedCentroid":
    #            Cats = np.zeros(len(X))
    #            if UsedNodes:
    #                Cats[UsedNodes] = WeiVal
    #
    #            p = PlotPG(X=X[Cats > 0, :], TargetPG=TargetPG, GroupsLab=Cats[Cats > 0])
    #            print(p)
    #
    #            p1 = PlotPG(X=X, TargetPG=TargetPG, GroupsLab=Cats)
    #            print(p1)

    return TargetPG


def CollapseBranches(
    X, PG, Mode="PointNumber", ControlPar=5, TrimmingRadius=float("inf")
):
    """
    Filter "small" branches

    X numeric matrix, the data matrix
    TargetPG list, the ElPiGraph structure to extend
    TrimmingRadius positive numeric, the trimming radius used to control distance
    ControlPar positive numeric, the paramter used to control the contribution of the different data points
    Mode string, the mode used to extend the graph. "PointNumber", "PointNumber_Extrema", "PointNumber_Leaves",
    "EdgesNumber", and "EdgesLength" are currently implemented
    PlotSelected boolean, should a diagnostic plot be visualized (currently not implemented)

    Return
    -------
    a list with 2 values: Nodes (a matrix containing the new nodes positions) and Edges (a matrix describing the new edge structure)

    The value of ControlPar has a different interpretation depending on the valus of Mode.

    If Mode = "PointNumber", branches with less that ControlPar points projected on the branch
    (points projected on the extreme points are not considered) are removed

    If Mode = "PointNumber_Extrema", branches with less that ControlPar points projected on the branch or the extreme
    points are removed

    If Mode = "PointNumber_Leaves", branches with less that ControlPar points projected on the branch and any leaf points
    (points projected on non-leaf extreme points are not considered) are removed

    If Mode = "EdgesNumber", branches with less that ControlPar edges are removed

    If Mode = "EdgesLength", branches with with a length smaller than ControlPar are removed
    """
    TargetPG = copy.deepcopy(PG)
    # Generate net
    Net = ConstructGraph(PrintGraph=TargetPG)

    # Set a color for the edges
    Net.es.set_attribute_values("status", "keep")

    # Get the leaves
    Leaves = np.where(np.array(Net.degree(mode="all")) == 1)[0]

    # get the partition
    PartStruct = PartitionData(
        X=X,
        NodePositions=TargetPG["NodePositions"],
        MaxBlockSize=100000000,
        TrimmingRadius=TrimmingRadius,
        SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
    )

    # Project points onto the graph
    ProjStruct = project_point_onto_graph(
        X=X,
        NodePositions=TargetPG["NodePositions"],
        Edges=TargetPG["Edges"][0],
        Partition=PartStruct[0],
    )

    # get branches
    Branches = GetSubGraph(Net=Net, Structure="branches")

    # get the number of points on the different branches
    AllBrInfo = []
    for BrNodes in Branches:
        # print(BrNodes)
        PotentialPoints = np.array([False] * len(ProjStruct["EdgeID"]))
        NodeNames = np.array(BrNodes)

        # Get the points on the extrema
        StartEdg = np.where(
            np.any(ProjStruct["Edges"] == NodeNames[0], axis=1)
        )[0]

        StartOnNode = np.array([False] * len(ProjStruct["EdgeID"]))

        SelPoints = np.isin(ProjStruct["EdgeID"], StartEdg)
        StartOnNode[SelPoints] = (
            ProjStruct["ProjectionValues"][SelPoints] > 1
        ) | (ProjStruct["ProjectionValues"][SelPoints] < 0)

        # EndEdg = np.where(np.any(ProjStruct["Edges"] == NodeNames[-1], axis=1))[0]
        EndOnNode = np.array([False] * len(ProjStruct["EdgeID"]))

        SelPoints = np.isin(ProjStruct["EdgeID"], EndOnNode)
        EndOnNode[SelPoints] = (
            ProjStruct["ProjectionValues"][SelPoints] > 1
        ) | (ProjStruct["ProjectionValues"][SelPoints] < 0)

        EdgLen = 0

        # Get the points on the branch (extrema are excluded)
        for i in range(1, len(BrNodes)):

            # Get the edge on the segment
            WorkingEdg = np.where(
                list(
                    map(
                        lambda x: all(
                            np.isin(x, NodeNames[range((i - 1), i + 1)])
                        ),
                        ProjStruct["Edges"],
                    )
                )
            )[0].squeeze()
            # Get the len of the segment
            EdgLen = EdgLen + ProjStruct["EdgeLen"][WorkingEdg]

            # Get the points on the segment
            Points = ProjStruct["EdgeID"] == WorkingEdg
            Points[np.isnan(Points)] = False

            # Is the edge in the right direction?
            if all(
                ProjStruct["Edges"][
                    WorkingEdg,
                ]
                == NodeNames[range((i - 1), i + 1)]
            ):
                Reverse = False
            else:
                Reverse = True

            # Counting points at the beginning

            if i == 1 and len(BrNodes) > 2:
                if Reverse:
                    PotentialPoints[Points] = (
                        ProjStruct["ProjectionValues"][Points] < 1
                    ) | PotentialPoints[Points]
                else:
                    PotentialPoints[Points] = (
                        ProjStruct["ProjectionValues"][Points] > 0
                    ) | PotentialPoints[Points]
                continue

            # Counting points at the end
            if i == (len(BrNodes) - 1):
                if Reverse:
                    PotentialPoints[Points] = (
                        ProjStruct["ProjectionValues"][Points] > 0
                    ) | PotentialPoints[Points]

                else:
                    PotentialPoints[Points] = (
                        ProjStruct["ProjectionValues"][Points] < 1
                    ) | PotentialPoints[Points]
                continue

            # all the other cases
            PotentialPoints[Points] = (
                (ProjStruct["ProjectionValues"][Points] > 0)
                & (ProjStruct["ProjectionValues"][Points] < 1)
            ) | PotentialPoints[Points]

        PointsOnEdgesLeaf = PotentialPoints

        if np.isin(NodeNames[0], Leaves):
            PointsOnEdgesLeaf = PointsOnEdgesLeaf | StartOnNode

        if np.isin(NodeNames[-1], Leaves):
            PointsOnEdgesLeaf = PointsOnEdgesLeaf | EndOnNode

        AllBrInfo.append(
            dict(
                PointsOnEdges=sum(PotentialPoints),
                PointsOnEdgeExtBoth=sum(
                    PotentialPoints | StartOnNode | EndOnNode
                ),
                PointsOnEdgesLeaf=sum(PointsOnEdgesLeaf),
                EdgesCount=len(BrNodes) - 1,
                EdgesLen=EdgLen,
            )
        )

    # Now all the information has been pre-computed and it is possible to filter

    if Mode == "PointNumber":
        ToFilter = (
            np.array([i["PointsOnEdges"] for i in AllBrInfo]) < ControlPar
        )

    if Mode == "PointNumber_Extrema":
        ToFilter = (
            np.array([i["PointsOnEdgeExtBoth"] for i in AllBrInfo])
            < ControlPar
        )

    if Mode == "PointNumber_Leaves":
        ToFilter = (
            np.array([i["PointsOnEdgesLeaf"] for i in AllBrInfo]) < ControlPar
        )

    if Mode == "EdgesNumber":
        ToFilter = np.array([i["EdgesCount"] for i in AllBrInfo]) < ControlPar

    if Mode == "EdgesLength":
        ToFilter = np.array([i["EdgesLen"] for i in AllBrInfo]) < ControlPar

    # Nothing to filter
    if sum(ToFilter) == 0:
        return dict(
            Edges=TargetPG["Edges"][0], Nodes=TargetPG["NodePositions"]
        )

    # TargetPG_New = TargetPG
    # NodesToRemove = None

    # Keep track of all the nodes to remove
    AllNodes_InternalBranches = {}
    # For all the branches
    for i in range(len(ToFilter)):

        # If we need to filter this
        if ToFilter[i] == True:

            NodeNames = np.array(Branches[i])

            # Is it a final branch ?
            if any(np.isin(NodeNames[[0, -1]], Leaves)):
                # It's a terminal branch, we can safely take it out

                print("Removing the terminal branch with nodes:", NodeNames)

                if len(NodeNames) > 2:
                    NodeNames_Ext = [NodeNames[0]]
                    NodeNames_Ext.extend(list(np.repeat(NodeNames[1:-1], 2)))
                    NodeNames_Ext.append(NodeNames[-1])
                else:
                    NodeNames_Ext = NodeNames
                # Set edges to be removed
                for e in range(0, len(NodeNames_Ext), 2):
                    rm_eid = Net.get_eid(
                        NodeNames_Ext[e], NodeNames_Ext[e + 1]
                    )
                    Net.es[rm_eid]["status"] = "remove"

            else:
                # It's a "bridge". We cannot simply remove nodes. Need to introduce a new one by "fusing" two stars
                print("Removing the bridge branch with nodes:", NodeNames)

                # Update the list of nodes to update
                AllNodes_InternalBranches = list(
                    set(AllNodes_InternalBranches).union(NodeNames)
                )

    # Create the network that will contain the final filtered network
    Ret_Net = copy.deepcopy(Net)

    # Get a net with all the groups of bridges to remove
    tNet = Net.induced_subgraph(AllNodes_InternalBranches)

    if tNet.vcount() > 0:
        # Get the different connected components
        CC = tNet.components()
        # Get the nodes associated with the connected components
        Member_Comps = CC.membership
        Vertex_Comps = [
            np.array(tNet.vs["name"])[np.where(Member_Comps == i)[0]]
            for i in np.unique(Member_Comps)
        ]
        # Get the centroid of the different connected components
        Centroids = np.array(
            [
                np.mean(TargetPG["NodePositions"][np.array(i)], axis=0)
                for i in Vertex_Comps
            ]
        )

        # Prepare a vector that will be used to contract vertices
        CVet = np.array(range(Net.vcount()))

        # For each centroid
        for i in range(len(Vertex_Comps)):
            # Add a new vertex
            Ret_Net.add_vertex(Ret_Net.vcount())
            # Add a new element to the contraction vector
            CVet = np.append(CVet, len(CVet))
            # specify the nodes that will collapse on the new node
            CVet[Vertex_Comps[i]] = len(CVet) - 1

        # collapse the network
        Ret_Net.contract_vertices(mapping=CVet)

    # delete edges belonging to the terminal branches
    edge_list = Ret_Net.get_edgelist()
    Ret_Net.delete_edges(
        [
            edge_list[i]
            for i in range(len(edge_list))
            if Ret_Net.es["status"][i] == "remove"
        ]
    )
    # remove loops that may have been introduced because of the collapse
    Ret_Net.simplify(loops=True)
    # # Remove empty nodes
    names = np.array(Ret_Net.vs.indices)[np.array(Ret_Net.degree()) > 0]
    Ret_Net = Ret_Net.induced_subgraph(
        np.array(Ret_Net.vs.indices)[np.array(Ret_Net.degree()) > 0]
    )

    if tNet.vcount() > 0:
        NodeMat = np.vstack((TargetPG["NodePositions"], Centroids))

    else:
        NodeMat = TargetPG["NodePositions"]

    NodeMat = NodeMat[names, :]

    return dict(Edges=np.array(Ret_Net.get_edgelist()), Nodes=NodeMat)


#' Title
#'
#' TargetPG
#' NodesToRemove
#'
#' @return
#'
#' @examples

# RemoveNodesbyIDs <- function(TargetPG, NodesToRemove) {

#   RemapNodeID <- cbind(
#     1:nrow(TargetPG$NodePositions),
#     1:nrow(TargetPG$NodePositions)
#   )

#   TargetPG_New <- TargetPG

#   # Remove nodes and edges
#   TargetPG_New$NodePositions <- TargetPG_New$NodePositions[-NodesToRemove, ]
#   TargetPG_New$ElasticMatrix <- TargetPG_New$ElasticMatrix[-NodesToRemove, -NodesToRemove]

#   # tEdges <- which(TargetPG_New_New$ElasticMatrix > 0, arr.ind = TRUE)
#   # tEdges <- tEdges[tEdges[,2] > tEdges[,1],]

#   # Remap Nodes IDs
#   RemapNodeID[RemapNodeID[,2] %in% NodesToRemove,2] <- 0
#   for(j in 1:nrow(RemapNodeID)){
#     if(RemapNodeID[j,2] == 0){
#       # the node has been removed. Remapping
#       RemapNodeID[RemapNodeID[,2] >= j, 2] <- RemapNodeID[RemapNodeID[,2] >= j, 2] - 1
#     }
#   }

#   tEdges <- TargetPG_New$Edges$Edges
#   for(j in 1:nrow(RemapNodeID)){
#     tEdges[TargetPG_New$Edges$Edges == RemapNodeID[j,1]] <- RemapNodeID[j,2]
#   }
#   EdgesToRemove <- which(rowSums(tEdges == 0) > 0)
#   tEdges <- tEdges[-EdgesToRemove, ]

#   TargetPG_New$Edges$Edges <- tEdges
#   TargetPG_New$Edges$Lambdas <- TargetPG_New$Edges$Lambdas[-EdgesToRemove]
#   TargetPG_New$Edges$Mus <- TargetPG_New$Edges$Mus[-NodesToRemove]

#   return(TargetPG_New)
# }


def ShiftBranching(
    X,
    PG,
    SelectionMode="NodePoints",
    DensityRadius=None,
    MaxShift=3,
    Compensate=False,
    BrIds=None,
    BrIdsTarget=None,
    TrimmingRadius=float("inf"),
):
    """
    Move branching nodes to areas with higher point density

    X numeric matrix, the data matrix
    TargetPG list, the ElPiGraph structure to extend
    TrimmingRadius positive numeric, the trimming radius used to control distance
    SelectionMode string, the mode to use to shift the branching points. The "NodePoints" and "NodeDensity" modes are currently supported
    DensityRadius positive numeric, the radius to be used when computing point density if SelectionMode = "NodeDensity"
    MaxShift positive integer, the maxium distance (as number of edges) to consider when exploring the branching point neighborhood
    Compensate booelan, should new points be included to compensate for olter one being removed (currently not implemented)
    BrIds integer vector, the id of the branching points to consider. Id not associted with node possessing degree > 2 will be ignored

    Return
    -------
    a list with two components: NodePositions (Containing the new nodes positions) and Edges (containing the new edges)

    The function explore the neighborhood of branching point for nodes with higher point density. It such point is found, to graph will be
    modified so that the new found node will be the new branching point of the neighborhood.

    @examples

    """

    TargetPG = copy.deepcopy(PG)

    Net = ConstructGraph(PrintGraph=TargetPG)
    BrPoints = np.where(np.array(Net.degree()) > 2)[0]

    if BrIds is None:
        BrIds = BrPoints
    else:
        BrIds = set(BrIds).intersection(BrPoints)

    PD = PartitionData(
        X=X,
        NodePositions=TargetPG["NodePositions"],
        MaxBlockSize=100000000,
        TrimmingRadius=TrimmingRadius,
        SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
    )

       
            
    for i,br in enumerate(BrIds):

        Neis = Net.neighborhood(br, order=MaxShift)
        # Neis = setdiff(as.integer(Neis), br)
        if SelectionMode == "Manual":
            if not(BrIdsTarget[i] in Neis):
                raise ValueError(f'New branching node is not a MaxShift={MaxShift} neighbor of the old branching node')
            NewBR = BrIdsTarget[i]
            
        else:
            if SelectionMode == "NodePoints":
                Neival = np.array(list(map(lambda x: np.sum(PD[0] == x), Neis)))

            if SelectionMode == "NodeDensity":
                Dists = PartialDistance(X, TargetPG["NodePositions"][Neis])

                if DensityRadius is None:
                    raise ValueError(
                        "DensityRadius needs to be specified when SelectionMode ="
                        " 'NodeDensity'"
                    )

                else:
                    Neival = np.sum(Dists < DensityRadius, axis=0)

            NeiDist = np.array(Net.shortest_paths(br, Neis, mode="all"))
            Neival = Neival[np.argsort(NeiDist, kind="mergesort")]
            NewBR = Neis[np.min(np.where(Neival.squeeze() == np.max(Neival))[0])]

        if NewBR != br:

            print("Moving the branching point at node", br)

            ToReconnect = Net.neighbors(br)
            # delete the edges forming the old branching point
            Net.delete_edges(Net.incident(br))

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # get paths from the new branching point to the old nodes
                AllPath = Net.get_shortest_paths(
                    NewBR, to=ToReconnect, output="epath"
                )
                AllPath_vpath = Net.get_shortest_paths(
                    NewBR, to=ToReconnect, output="vpath"
                )

            # delete these extra nodes
            for i in range(len(AllPath)):
                if len(AllPath[i]) > 0:
                    Net.delete_edges(AllPath[i])

            # reconnect the old nodes to the new branching points, excep for any node found in the previuos operation
            ToReconnect = list(
                set(ToReconnect).difference(
                    [j for i in AllPath_vpath for j in i]
                )
            )
            Net.add_edges([(NewBR, i) for i in ToReconnect])

    if Compensate:
        print("Warning : Node compensation is not implemented yet")

    NewNodePositions = TargetPG["NodePositions"][np.array(Net.degree()) > 0, :]

    Net.delete_vertices(np.where(np.array(Net.degree()) == 0)[0])
    NewEdges = np.array(Net.get_edgelist())

    return dict(NodePositions=NewNodePositions, Edges=NewEdges)


#' Filter cliques
#'
#' TargetPG
#' MaxClZize
#' DistThr
#'
#' @return
#'
#'
#' @examples
# CollapseCliques <- function(TargetPG, MaxClZize = NULL, DistThr = NULL, Compensate = 3) {

#   Net <- ConstructGraph(TargetPG)

#   Nodes <- TargetPG$NodePositions

#   Cliques <- igraph::cliques(Net, min = 3, max = MaxClZize)

#   ClSize <- sapply(Cliques, length)

#   print(paste(length(ClSize), "cliques detected"))

#   Tries <- 0

#   while (TRUE) {

#     if(Tries > 100 | length(ClSize) == 0){
#       return(list(
#         Nodes = TargetPG$NodePositions,
#         Edges = TargetPG$Edges$Edges
#       ))
#     }

#     Tries <- Tries + 1

#     ClToCollapse <- sample(x = which(ClSize == max(ClSize)), size = 1)
#     NodesToCollapse <- as.integer(Cliques[[ClToCollapse]])

#     NewNode <- colMeans(Nodes[NodesToCollapse, ])

#     if(!is.null(DistThr)){
#       if(mean(distutils::PartialDistance(matrix(NewNode, nrow = 1), Nodes[NodesToCollapse, ])) < DistThr){
#         break()
#       }
#     } else {
#       break()
#     }

#   }

#   Net <- igraph::add.vertices(Net, 1, attr = list(name = paste(igraph::vcount(Net)+1)))

#   ContractNet <- 1:igraph::vcount(Net)
#   ContractNet[NodesToCollapse] <- igraph::vcount(Net)

#   ContrNet <- igraph::contract.vertices(Net, ContractNet)
#   igraph::V(ContrNet)$name <- sapply(igraph::V(ContrNet)$name, function(x){
#     if(length(x)>0){
#       max(as.numeric(x))
#     } else {
#       0
#     }
#   })
#   ContrNet <- igraph::delete.vertices(ContrNet, NodesToCollapse)
#   ContrNet <- igraph::simplify(ContrNet, remove.loops = TRUE, remove.multiple = TRUE)
#   # igraph::V(ContrNet)$name <- paste(1:igraph::vcount(ContrNet))

#   UpdateNodes <- rbind(Nodes, NewNode)[as.integer(igraph::V(ContrNet)$name),]
#   rownames(UpdateNodes) <- NULL

#   if(Compensate <= max(ClSize)){
#     Neis <- unlist((igraph::adjacent_vertices(graph = ContrNet, v = paste(igraph::vcount(Net)))[[1]])$name)

#     NNodes <- t((t(Nodes[Neis, ]) + NewNode)/2)

#     ContrNet <- igraph::add.vertices(ContrNet, length(Neis), attr = list(name = paste0("Ext_", Neis)))
#     ContrNet <- igraph::delete_edges(ContrNet, igraph::incident_edges(graph = ContrNet, v = paste(igraph::vcount(Net)))[[1]])
#     ContrNet <- igraph::add.edges(graph = ContrNet, edges = rbind(paste(igraph::vcount(Net)), paste0("Ext_", Neis)))
#     ContrNet <- igraph::add.edges(graph = ContrNet, edges = rbind(paste(Neis), paste0("Ext_", Neis)))

#     UpdateNodes <- rbind(UpdateNodes, NNodes)

#   }

#   return(list(
#     Nodes = UpdateNodes,
#     Edges = igraph::get.edgelist(ContrNet, names = FALSE)
#   ))

# }


#'
#'
#' #' Title
#' #'
#' #' variables
#' #'
#' #' @return
#' #'
#' #'
#' #' @examples
#' RewireBranches <- function(X,
#'                        TargetPG,
#'                        MaxDist = 3,
#'                        Lambda,
#'                        Mu,
#'                        MaxNumberOfIterations = 10,
#'                        eps = 0.01,
#'                        FinalEnergy = "Base",
#'                        alpha = 0,
#'                        beta = 0,
#'                        Mode = 1,
#'                        TrimmingRadius = Inf,
#'                        FastSolve = FALSE,
#'                        prob = 1) {
#'
#'   TargetPG1 <- TargetPG
#'
#'   NodeDist <- distutils::PartialDistance(TargetPG$NodePositions, TargetPG$NodePositions)
#'
#'   Net <- ConstructGraph(PrintGraph = TargetPG1)
#'   BrPoint <- which(igraph::degree(Net)>2)
#'
#'   for(i in 1:length(BrPoint)){
#'
#'     Net <- ConstructGraph(PrintGraph = TargetPG1)
#'
#'     NeiBr <- as.integer(names(igraph::neighborhood(graph = Net, order = 1, nodes = BrPoint[i])[[1]]))
#'     NeiBr <- setdiff(NeiBr, BrPoint[i])
#'
#'     for(j in 1:length(NeiBr)){
#'       TargetPG1$NodePositions <- rbind(
#'         TargetPG1$NodePositions,
#'         colMeans(
#'           rbind(
#'             TargetPG$NodePositions[BrPoint[i], ],
#'             TargetPG$NodePositions[NeiBr[j], ]
#'           )
#'         )
#'       )
#'       OldEdg <- TargetPG1$Edges$Edges[,1] %in% c(BrPoint[i], NeiBr[j]) &
#'         TargetPG1$Edges$Edges[,2] %in% c(BrPoint[i], NeiBr[j])
#'       TargetPG1$Edges$Edges <- TargetPG1$Edges$Edges[!OldEdg,]
#'       TargetPG1$Edges$Edges <- rbind(TargetPG1$Edges$Edges,
#'                                      rbind(
#'                                        c(BrPoint[i], nrow(TargetPG1$NodePositions)),
#'                                        c(NeiBr[j], nrow(TargetPG1$NodePositions))
#'                                      )
#'       )
#'
#'     }
#'
#'     Net <- ConstructGraph(PrintGraph = TargetPG1)
#'     NeiBr <- as.integer(names(igraph::neighborhood(graph = Net, order = 1, nodes = BrPoint[i])[[1]]))
#'     NeiBr <- setdiff(NeiBr, BrPoint[i])
#'
#'     StarEdes <- igraph::get.edge.ids(graph = Net, vp = rbind(rep(BrPoint[i], length(NeiBr)), NeiBr), directed = FALSE)
#'
#'     tNet <- igraph::delete.edges(graph = Net, edges = StarEdes)
#'     Neigh <- igraph::neighborhood(graph = tNet, order = igraph::vcount(tNet), nodes = NeiBr)
#'     NeighDist <- lapply(Neigh, function(NeiVect){
#'       sapply(igraph::shortest_paths(graph = Net, from = BrPoint[i], to = as.vector(NeiVect))$vpath, length)
#'     })
#'
#'     Mats <- list()
#'
#'     for(j in 1:length(Neigh)){
#'
#'       tNet <- igraph::delete.edges(graph = Net, igraph::E(Net)[igraph::`%--%`(BrPoint[i], NeiBr[j])])
#'
#'       VertexToTest <- Neigh[-j]
#'       VertexToTest.Dist <- NeighDist[-j]
#'
#'       VertexToTest <- lapply(1:length(VertexToTest), function(i){
#'         (VertexToTest[[i]])[VertexToTest.Dist[[i]] <= MaxDist]
#'       })
#'       VertexToTest <- unlist(VertexToTest)
#'
#'       for(k in 1:length(VertexToTest)){
#'         TestNet <- igraph::add.edges(graph = tNet, edges = c(NeiBr[j], VertexToTest[k]))
#'         Mats[[length(Mats)+1]] <- apply(igraph::get.edgelist(TestNet), 2, as.numeric)
#'       }
#'
#'     }
#'
#'     SquaredX <- rowSums(X^2)
#'
#'     Embed <- lapply(Mats, function(mat){
#'       PrimitiveElasticGraphEmbedment(X = X,
#'                                                    NodePositions = TargetPG1$NodePositions,
#'                                                    ElasticMatrix = Encode2ElasticMatrix(mat, Lambda, Mu),
#'                                                    SquaredX = SquaredX,
#'                                                    verbose = FALSE,
#'                                                    MaxNumberOfIterations = MaxNumberOfIterations,
#'                                                    eps = eps,
#'                                                    FinalEnergy = FinalEnergy,
#'                                                    alpha = alpha,
#'                                                    beta = beta,
#'                                                    Mode = Mode,
#'                                                    TrimmingRadius = TrimmingRadius,
#'                                                    FastSolve = FastSolve,
#'                                                    prob = prob)
#'     })
#'
#'
#'     # SumDists <- sapply(Mats, function(EdgMat){
#'     #   ProjStruct <- project_point_onto_graph(X, NodePositions = TargetPG$NodePositions, Edges = EdgMat)
#'     #   OnEdes <- ProjStruct$ProjectionValues > 0 & ProjStruct$ProjectionValues < 1
#'     #   DistFromEdge <- rowSums((ProjStruct$X_projected - X)^2)
#'     #   c(sum(DistFromEdge[OnEdes]), sum(ProjStruct$EdgeLen))
#'     # })
#'
#'     # order(colSums(SumDists))
#'     # which.min(SumDists[2,])
#'
#'
#'     # TargetPG2 <- TargetPG1
#'
#'     TargetPG1$Edges$Edges <- Mats[[which.min(sapply(Embed, "[[", "ElasticEnergy"))]]
#'
#'     PlotPG(X, TargetPG1, DimToPlot = 1:3, NodeLabels = 1:nrow(TargetPG1$NodePositions), LabMult = 3)
#'
#'     which(igraph::degree(ConstructGraph(TargetPG1))>2)
#'
#'   }
#'
#'   #
#'   #
#'   #
#'   # MaxByEdg <- aggregate(DistFromEdge, by=list(ProjStruct$EdgeID), max)
#'   # ToMove <- MaxByEdg[which.max(MaxByEdg[,2]),1]
#'   #
#'   # Ends <- TargetPG$Edges$Edges[ToMove,]
#'   #
#'   #
#'   # Net <-
#'   # igraph::neighborhood(graph = Net, order = igraph::vcount(Net), nodes = Ends)
#'
#'   return(TargetPG1)
#'
#' }
#'
#'
#'
#'
